"""
The _compat module is used for code which requires branching between different
Python environments. It is excluded from the code coverage checks.
"""
import asyncio
import ssl
import sys
import typing

if typing.TYPE_CHECKING:  # pragma: no cover
    import typing_extensions

# Brotli support is optional
# The C bindings in `brotli` are recommended for CPython.
# The CFFI bindings in `brotlicffi` are recommended for PyPy and everything else.
try:
    import brotlicffi as brotli
except ImportError:  # pragma: nocover
    try:
        import brotli
    except ImportError:
        brotli = None

if sys.version_info >= (3, 10) or (
    sys.version_info >= (3, 7) and ssl.OPENSSL_VERSION_INFO >= (1, 1, 0, 7)
):

    def set_minimum_tls_version_1_2(context: ssl.SSLContext) -> None:
        # The OP_NO_SSL* and OP_NO_TLS* become deprecated in favor of
        # 'SSLContext.minimum_version' from Python 3.7 onwards, however
        # this attribute is not available unless the ssl module is compiled
        # with OpenSSL 1.1.0g or newer.
        # https://docs.python.org/3.10/library/ssl.html#ssl.SSLContext.minimum_version
        # https://docs.python.org/3.7/library/ssl.html#ssl.SSLContext.minimum_version
        context.minimum_version = ssl.TLSVersion.TLSv1_2

else:

    def set_minimum_tls_version_1_2(context: ssl.SSLContext) -> None:
        # If 'minimum_version' isn't available, we configure these options with
        # the older deprecated variants.
        context.options |= ssl.OP_NO_SSLv2
        context.options |= ssl.OP_NO_SSLv3
        context.options |= ssl.OP_NO_TLSv1
        context.options |= ssl.OP_NO_TLSv1_1


def iscoroutine(coro: object) -> "typing_extensions.TypeGuard[typing.Coroutine]":
    # Drop when this is resolved: https://github.com/python/typeshed/pull/8104
    return asyncio.iscoroutine(coro)



""" SQLAlchemy Manager

Column:
    * primary_key
    * nullable
    * index
    * unique
"""
import functools
import typing
from collections import OrderedDict, namedtuple

from .. import manager
from ..core.custom_class import Globals

try:
    import sqlalchemy
    from sqlalchemy.orm import declarative_base

    SQLALCHEMY_ACTIVE = True
    SA = sqlalchemy
except ImportError:
    SQLALCHEMY_ACTIVE = False
    SA = None
    declarative_base = None

try:
    from databases import Database
except ImportError:
    Database = None

DatabaseManager = namedtuple("SQL", ["base", "database", "controller"])

# CORE Fields
ALL_FIELDS = [
    "String",
    "Text",
    "Integer",
    "Float",
    "Boolean",
    "Date",
    "DateTime",
    "Time",
    "Decimal",
    "ID",
    "TEXT",
    "JSON",
]

# Init Fields
Scalar = {key: key for key in ALL_FIELDS}

# Map SQL
if SQLALCHEMY_ACTIVE:
    # Column = SA.Column(SA.Integer, primary_key=True)
    # Fields
    Scalar["String"] = SA.String(length=255)
    Scalar["Integer"] = SA.Integer
    Scalar["Float"] = SA.Float
    Scalar["Boolean"] = SA.Boolean
    Scalar["Date"] = SA.Date
    Scalar["DateTime"] = SA.DateTime
    Scalar["Time"] = SA.Time
    Scalar["Decimal"] = SA.String(length=255)
    # Custom
    Scalar["ID"] = SA.Integer
    Scalar["Text"] = SA.Text
    Scalar["JSON"] = SA.JSON


def column_base_setup(name: str, config: dict) -> dict:
    """Column Base"""
    return {
        "primary_key": name in config.get("primary_key", []),
        "index": name in config.get("index", []),
        "unique": name in config.get("unique", []),
        "nullable": not name in config.get("required", []),
    }


def column_setup(*fields, name: str = None, config: dict = None):
    """Column Setup"""
    setup = column_base_setup(name, config)
    return SA.Column(*fields, **setup)


def sql_fields(annotations, config) -> OrderedDict:
    """All Column(s) Setup"""
    db_fields = OrderedDict()
    related_fields = OrderedDict()
    for field in annotations:
        column = functools.partial(column_setup, name=field.name, config=config)
        if field.is_custom_type:
            my_class = Globals.get(field.real)
            class_config = my_class.__spoc__.config
            related = class_config["table_name"]
            match class_config["engine"]:
                case "sql":
                    sql_scalar = column(SA.Integer, SA.ForeignKey(f"{related}._id"))
                case _:
                    sql_scalar = column(SA.String(length=255))
            if field.name not in config["ignore"]:
                related_fields[f"{field.name}_id"] = sql_scalar
        else:
            if field.name not in config["ignore"]:
                db_fields[field.name] = column(Scalar.get(field.scalar))
    for key, val in related_fields.items():
        db_fields[key] = val
    return db_fields


def sql_orm(
    base: typing.Any = None,
    original_object: typing.Any = None,
    setup: typing.Any = None,
):
    """SQLAlchemy ORM Setup"""
    return type(f"SQL{original_object.__name__}", (base,), setup).__table__


def sql_model(
    controller: typing.Any = None,
    original_object: typing.Any = None,
    annotations: typing.Any = None,
    config: dict = None,
    new_object: typing.Any = None,
):
    """Model Setup"""
    sqlalchemy_class_setup = {
        "__tablename__": config["table_name"],
        "__table_args__": [],
        "_id": SA.Column(SA.Integer, primary_key=True),
    }
    sql_cols = sql_fields(annotations, config)
    for setup in config["unique_together"]:
        unique_together = SA.UniqueConstraint(*setup)
        sqlalchemy_class_setup["__table_args__"].append(unique_together)

    sqlalchemy_class_setup["__table_args__"] = tuple(
        sqlalchemy_class_setup["__table_args__"]
    )
    for key, val in sql_cols.items():
        if key != "id":
            sqlalchemy_class_setup[key] = val
    objects = sql_orm(
        controller.base, original_object=original_object, setup=sqlalchemy_class_setup
    )
    # Finally
    new_object.__database__ = "sql"
    new_object.objects = controller.controller(objects)
    return new_object


def create_sql_manager(url: str):
    """Create SQL Base-Manager"""
    base = declarative_base()
    database = Database(url)
    controller = functools.partial(manager.SQL, database=database)
    return DatabaseManager(base=base, database=database, controller=controller)



"""
    Tool to Transform { Text } to { Code-Cases }
    Tool to Get Module's Name
"""
import re

from .coro import coro
from .forms import Item, Pagination


def doc(description):
    """Inject Documentation"""

    def decorator(function):
        """Decorator"""
        function.__doc__ = description
        return function

    return decorator


def to_kebab_case(text):
    """Clean Text"""
    text = re.sub("[^0-9a-zA-Z]+", "-", str(text))
    return text.lower()


def to_camel_case(text):
    """Converts to { camelCase }"""
    text = to_kebab_case(text)
    init, *temp = text.title().split("-")
    return "".join([init.lower()] + temp)


def to_pascal_case(text):
    """Converts to { PascalCase }"""
    text = to_kebab_case(text)
    return text.title().replace("-", "")



"""
    Annotations
"""

import dataclasses as dc
import typing
from collections import OrderedDict

from .scalars import JSON, TEXT, Scalar

CUSTOM_SCALARS = [TEXT, JSON]


@dc.dataclass(frozen=True)
class AnnotationInfo:
    """Annotation Setup"""

    name: str
    scalar: str
    graphql: typing.Any
    original: typing.Any
    real: typing.Any
    optional: bool = False
    is_list: bool = False
    is_json: bool = False
    is_custom_type: bool = False


@dc.dataclass(frozen=True)
class Model:
    """Annotation Setup"""

    fields: typing.Any
    methods: typing.Any
    annotations: typing.Any


def make_annotation(
    real: typing.Any = None,
    is_list: bool = False,
    is_custom_type: bool = False,
):
    """Make Annotations"""

    col_type = real

    if is_custom_type:
        col_type = typing.ForwardRef(col_type)
    if is_list:
        col_type = typing.List[col_type]

    col_type = typing.Optional[col_type]

    return col_type


def get_field_annotation_core(single_annotation: typing.Any, found_args: tuple):
    """Field Core Annotation"""
    core_arg = None
    has_children = []

    # Core Arg
    if found_args:
        core_arg = found_args[0]
    else:
        core_arg = single_annotation

    # Core Arg Continued . . .
    has_children = typing.get_args(core_arg)
    if len(has_children) > 0:
        core_arg = has_children[0]
        if isinstance(core_arg, str):
            core_arg = typing.ForwardRef(core_arg)
    return core_arg


def get_annotations_args(field_name: str, single_annotation: typing.Any):
    """Get: Custom Annotations"""
    found_args = typing.get_args(single_annotation)
    real_arg = get_field_annotation_core(single_annotation, found_args)
    is_custom_type = isinstance(real_arg, typing.ForwardRef)
    is_optional = type(None) in found_args
    is_list = False
    is_dict = False
    scalar_name = None

    if isinstance(real_arg, str):
        is_custom_type = True

    if isinstance(real_arg, typing.ForwardRef):
        real_arg = real_arg.__forward_arg__

    scalar_field = Scalar.get(real_arg)
    if real_arg in CUSTOM_SCALARS:
        real_arg = scalar_field.python

    # IS -> List <YES or NO>
    if found_args and len(typing.get_args(found_args[0])) > 0:
        is_list = True

    # IS -> Step 1 ...
    if len(found_args) == 1:
        is_list = True

    # IS -> Step 2 ...
    if real_arg == JSON:
        is_dict = True

    # IS -> Step 3 ...
    if scalar_field:
        scalar_name = scalar_field.name

    # IS -> Step 4 ...
    graphql = make_annotation(
        real_arg,
        is_list=is_list,
        is_custom_type=is_custom_type,
    )

    # Custom Annotations
    return AnnotationInfo(
        name=field_name,
        scalar=scalar_name,
        original=single_annotation,
        real=real_arg,
        optional=is_optional,
        is_list=is_list,
        is_json=is_dict,
        is_custom_type=is_custom_type,
        graphql=graphql,
    )


def get_args(original_object):
    """Get Annotations"""
    real_annotations = original_object.__annotations__
    field_keys = list(real_annotations.keys())
    func_annotations = [
        x for x in dir(original_object) if not x.startswith("_") and x not in field_keys
    ]
    cool_annotations = OrderedDict()
    for key, val in real_annotations.items():
        found_args = get_annotations_args(key, val)
        cool_annotations[key] = found_args
    for key in func_annotations:
        func = getattr(original_object, key)
        if func:
            val = func.__annotations__.get("return")
            found_args = get_annotations_args(key, val)
            cool_annotations[key] = found_args
    return Model(
        fields=field_keys, methods=func_annotations, annotations=cool_annotations
    )


# is_list
# is_json
# optional
# is_custom_type



"""
    Fastberry Main (Core-Methods)
"""
from pathlib import Path

from starlette.middleware.base import BaseHTTPMiddleware as BaseMiddleware
from strawberry.extensions import Extension as BaseExtension
from strawberry.permission import BasePermission

from .scripts import templates_code

try:
    import spoc
except:
    if not Path("./config").exists():
        Path("./config").mkdir()

    file_settings = "./config/settings.py"
    if not Path(file_settings).exists():
        with open(file_settings, "w", encoding="utf-8") as f:
            f.write(templates_code.SETTINGS)

    file_settings = "./config/spoc.toml"
    if not Path(file_settings).exists():
        with open(file_settings, "w", encoding="utf-8") as f:
            f.write(templates_code.SPOC)

    file_settings = "./config/docs.md"
    if not Path(file_settings).exists():
        with open(file_settings, "w", encoding="utf-8") as f:
            f.write(templates_code.DOCS)

    import spoc

# FrameWork
from .components import APIRouter as Router
from .components import cli
from .components import graphql as gql
from .framework import Fastberry as App

# GraphQL Tools
from .graphql import edges, error, errors, mutation, page, query

# GraphQL Premade User-Inputs
from .tools import Item as item
from .tools import Pagination as pagination
from .tools import coro, doc

# Framework Wrappers
base_dir = spoc.base_dir
config = spoc.config
mode = spoc.mode
project = spoc.project
settings = spoc.settings

# Tools
component = spoc.component

try:
    import dbcontroller as dbc
    from dbcontroller.forms import ISNULL

    if hasattr(settings, "DATABASES"):
        config_sql = settings.DATABASES.get("sql")
        config_mongo = settings.DATABASES.get("mongo")
        default_sql = config_sql.get("default")
        default_mongo = config_mongo.get("default")
        if default_sql:
            sql = dbc.Controller(sql=default_sql)
        if default_mongo:
            mongo = dbc.Controller(mongo=default_mongo)

    # Types
    type = dbc.type

    # Forms
    input = dbc.form.graphql
    value = dbc.form.field

    # Value Tool
    filters = dbc.form.filters

    # Types Tool (DBController)
    field = dbc.field
    manager = dbc.manager

    # Scalars
    ID = dbc.ID
    date = dbc.date
    datetime = dbc.datetime
    time = dbc.time
    decimal = dbc.decimal
    text = dbc.text
    time = dbc.time
    json = dbc.json

    # Tester
    Date = dbc.Date

except ImportError:
    import strawberry

    # Types
    type = strawberry.type
    input = strawberry.input



"""
    Strawberry GraphQL Made Easy
"""
from .types import Edges as edges
from .types import Error as errors
from .types import ErrorMessage as error
from .types import Mutation as mutation
from .types import Page as page
from .types import Query as query



import asyncio
import functools
import typing


def is_async_callable(obj: typing.Any) -> bool:
    while isinstance(obj, functools.partial):
        obj = obj.func

    return asyncio.iscoroutinefunction(obj) or (
        callable(obj) and asyncio.iscoroutinefunction(obj.__call__)
    )



"""
    Form Creator
        Modules:
            * dataclass: Create a Custom DataClass.
            * Field: Create a Field for the dataclass.
"""
import dataclasses as dc
import functools
import re
import typing
from types import SimpleNamespace

from .core.annotations import get_args
from .core.spoc import COMPONENT, component
from .tools import to_pascal_case

try:
    import strawberry

    STRAWBERRY_INPUT = strawberry.input
except ImportError:
    STRAWBERRY_INPUT = False
    strawberry = None

# Custom Typing
ISNULL = typing.TypeVar("ISNULL", bool, None)


def get_custom_annotations(original_object):
    """Get Custom Annotations"""
    good = get_args(original_object)
    out_dict = {}
    for field in good.annotations.values():
        the_type = field.real
        if field.is_list:
            the_type = list
        if strawberry and the_type == strawberry.ID:
            the_type = str
        out_dict[field.name] = the_type
    return out_dict


@dc.dataclass
class FormField:
    """DataClass -> Field"""

    type: typing.Any
    name: str = None
    default: typing.Any = ISNULL
    required: bool = False
    regex: dict = dc.field(default_factory=dict)
    rules: list = dc.field(default_factory=list)
    filters: dict = dc.field(default_factory=dict)
    # fixed: bool = False


@dc.dataclass
class FormError:
    """DataClass -> Form Error"""

    field: str
    type: str
    text: str = None


@dc.dataclass
class FormResponse:
    """DataClass -> Form Response"""

    data: dc.field(default_factory=dict)
    errors: dc.field(default_factory=list)
    is_valid: bool = False
    next: typing.Any = None


@dc.dataclass
class FormCRUD:
    """DataClass -> Form Error"""

    create: typing.Any = None
    read: typing.Any = None
    update: typing.Any = None
    delete: typing.Any = None


def regex_search(pattern: str, text: str):
    """Search for { Regular Expression } Pattern"""

    # regex_search("[\w\.-]+@[\w\.-]+", "admin@example.com")
    def is_regex(x):
        "Real Regex"
        return x is not None

    return is_regex(re.search(pattern, str(text)))


def regex_replace(items: list[tuple], text: str):
    """Replace Text"""
    for selector, replacement in items:
        text = re.sub(selector, replacement, text)
    return text


def default_value(default):
    """Create Default Value"""
    is_callable = False
    if callable(default):
        is_callable = True
        field = dc.field(default_factory=default)
    else:
        field = dc.field(default=default)
    return is_callable, field


def custom_field_maker(info: FormField):
    """Custom Field Maker"""
    if info.default != ISNULL and not info.required:
        is_callable, default = default_value(info.default)
        if is_callable:
            the_type = typing.Optional[info.type]
        else:
            the_type = info.type
        field = tuple([info.name, the_type, default])
    else:
        extra_fields = []
        if not info.required:
            the_type = typing.Optional[info.type]
            extra_fields.extend([the_type, dc.field(default=None)])
        else:
            the_type = info.type
            extra_fields.append(the_type)
        field = tuple([info.name, *extra_fields])
    return field


class FormBase:
    """(Singleton) Create a Form"""

    def __new__(cls):
        """Class Starter"""
        it_id = "__it__"
        it = cls.__dict__.get(it_id, None)
        if it is not None:
            return it
        it = object.__new__(cls)
        setattr(cls, it_id, it)
        it._init_only_once_for_the_whole_class(cls)
        return it

    def _init_only_once_for_the_whole_class(self, cls):
        """Class __init__ Replacement"""
        builtin_keys = ["_init_only_once_for_the_whole_class", "_validate", "Next"]
        fields = [
            x for x in dir(self) if not x.startswith("__") and x not in builtin_keys
        ]
        validate = {}
        annotations = {}
        custom_annotations = {"one": [], "two": []}
        for f_name in fields:
            current = getattr(self, f_name)
            current.name = f_name
            validate[f_name] = current
            if current.required:
                annotations[f_name] = current.type
            else:
                annotations[f_name] = typing.Optional[current.type]
            # Custom Type Annotation
            custom_field = custom_field_maker(current)
            if not current.required:
                custom_field = (
                    custom_field[0],
                    typing.Optional[custom_field[1]],
                    custom_field[2],
                )
                custom_annotations["two"].append(custom_field)
            else:
                custom_annotations["one"].append(custom_field)
            # custom_annotations.append(custom_field)

        self._config = validate
        cls.__annotations__ = annotations
        cls.__custom_annotations__ = custom_annotations

    def __call__(self, form: dict = None):
        form = form or {}
        errors = []
        output_form = {**form}
        # Check Fields
        for name, setup in self._config.items():
            current_input = form.get(name)
            is_valid_type = False
            type_origin = typing.get_origin(setup.type)
            if type_origin is None:
                type_to_check = setup.type
            else:
                type_to_check = type_origin
            if strawberry and type_to_check == strawberry.ID:
                is_valid_type = True
            else:
                is_valid_type = isinstance(current_input, type_to_check)
            if not is_valid_type and current_input:
                the_error = FormError(
                    field=name, type="typing", text=f"{ type_to_check } is required."
                )
                errors.append(the_error.__dict__)
            # Required Validator
            if setup.required:
                the_error = FormError(
                    field=name, type="required", text=f"{ name } is required."
                )
                if name not in form.keys() or not current_input:
                    errors.append(the_error.__dict__)
            # Custom Validators
            if current_input:
                # Regex Validator
                for test, error_message in setup.regex.items():
                    found_regex = not regex_search(test, current_input)
                    if found_regex:
                        the_error = FormError(
                            field=name, type="regex", text=error_message
                        )
                        errors.append(the_error.__dict__)
                # Custom Rules Validator
                for test in setup.rules:
                    if callable(test):
                        try:
                            found_rule = test(current_input)
                            if found_rule is not True:
                                the_error = FormError(
                                    field=name, type="rule", text=found_rule
                                )
                                errors.append(the_error.__dict__)
                        except:
                            the_error = FormError(
                                field=name, type="except", text="invalid input"
                            )
                            errors.append(the_error.__dict__)
                # Custom Filters
                if len(errors) == 0:
                    regex_methods = []
                    new_input = current_input
                    if setup.filters:
                        for test in setup.filters.get("regex"):
                            if isinstance(test, list | tuple):
                                regex_methods.append(test)
                            if (
                                isinstance(current_input, str)
                                and len(regex_methods) > 0
                            ):
                                new_input = regex_replace(regex_methods, current_input)
                        for test in setup.filters.get("rules"):
                            if callable(test):
                                try:
                                    new_input = test(new_input)
                                except:
                                    pass
                    # Add Field To Form
                    output_form[name] = new_input
            # Set Default
            if not current_input:
                output_form[name] = setup.default

        if "input" in output_form:
            del output_form["input"]

        return FormResponse(
            data=SimpleNamespace(**output_form),
            errors=errors,
            is_valid=len(errors) == 0,
        )


def run_validator(form, validator):
    """Return Custom Response to the Dataclass"""
    user_input = validator(form.__dict__)
    if hasattr(validator, "Next"):
        if hasattr(validator.Next, "run"):
            if user_input.is_valid:
                user_input.next = validator.Next.run(user_input.data)
    form.input = user_input


def make_dataclass(BaseClass, form_name):
    """(Custom) Make Dataclass"""
    validator = BaseClass()
    form_annotations = BaseClass.__custom_annotations__
    cleaned_data = custom_field_maker(
        SimpleNamespace(name="input", type=typing.Any, default=None, required=False)
    )
    form_annotations["two"].append(cleaned_data)
    class_annotations = form_annotations["one"]
    class_annotations.extend(form_annotations["two"])
    OutClass = dc.make_dataclass(
        form_name,
        class_annotations,
        namespace={"__post_init__": lambda self: run_validator(self, validator)},
    )
    del OutClass.__annotations__["input"]
    return OutClass


def dataclass(
    original_object: object = None,
    *,
    name: str = None,
    prefix: str | list[str] = None,
    suffix: str | list[str] = None,
    graphql: bool = True,
    description: str = None,
):
    """Form To GQL Input"""

    # Starting Wrapper. . .
    if original_object is None:
        return functools.partial(
            dataclass,
            name=name,
            prefix=prefix,
            suffix=suffix,
            graphql=graphql,
            description=description,
        )

    # Configure Class
    if name:
        form_name = name
    else:
        form_name = ""
        # Class Prefix
        if prefix:
            if isinstance(prefix, list):
                for fix in prefix:
                    form_name += fix.title()
            else:
                form_name += prefix.title()
        # Class Name
        form_name += f"{original_object.__name__}"
        # Class Suffix
        if suffix:
            if isinstance(suffix, list):
                for fix in suffix:
                    form_name += fix.title()
            else:
                form_name += suffix.title()
    # Re-Create Class with Form
    custom_class = type(original_object.__name__, (FormBase, original_object), {})

    # Create Data-Class
    OutClass = make_dataclass(custom_class, form_name)
    if STRAWBERRY_INPUT and graphql:
        description = description or original_object.__doc__
        OutClass = STRAWBERRY_INPUT(OutClass, description=description)

    # Create Component
    out_dict = get_custom_annotations(OutClass)
    component(OutClass, config={"annotations": out_dict}, metadata=COMPONENT["form"])
    return OutClass


def form_filters(
    regex: list[tuple[str, str]] = None, rules: list[typing.Callable] = None
) -> dict:
    """Form Filters"""
    regex = regex or []
    rules = rules or []
    return {
        "regex": regex,
        "rules": rules,
    }


def form_crud(model: str = None) -> FormCRUD:
    """Form CRUD"""
    if not model:
        raise ValueError("Model must be specified.")

    def form_crud_read(search_name: str = None):
        """Form CRUD: <Read>"""
        name = model
        if search_name:
            name += to_pascal_case(search_name)
        return functools.partial(dataclass, name=f"{name}ReadCRUD")

    return FormCRUD(
        create=functools.partial(dataclass, name=f"{model.title()}CreateCRUD"),
        read=form_crud_read,
        update=functools.partial(dataclass, name=f"{model.title()}UpdateCRUD"),
        delete=functools.partial(dataclass, name=f"{model.title()}DeleteCRUD"),
    )


class Form:
    """Form Wrapper"""

    field = FormField
    filters = form_filters

    # Custom DataClasses
    dataclass = dataclass
    crud = form_crud

    @staticmethod
    def group(
        prefix: str | list[str] = None,
        suffix: str | list[str] = None,
    ):
        """Form(s) with a { Suffix } or { Prefix } in Common"""
        return functools.partial(dataclass, prefix=prefix, suffix=suffix, graphql=False)

    @staticmethod
    def graphql(
        prefix: str | list[str] = None,
        suffix: str | list[str] = None,
    ):
        """GraphQL-Form(s) with a { Suffix } or { Prefix } in Common"""
        return functools.partial(dataclass, prefix=prefix, suffix=suffix, graphql=True)



""" [Complex-Tools]
"""
from .schema_reader import INTROSPECTION_QUERY, get_info


def introspection_info(schema):
    """GraphQL Introspection INFO"""
    result = schema.execute_sync(INTROSPECTION_QUERY)
    return get_info(result.data)



"""[summary]
Mongo & SQL - Object-Manager (Utils)
"""

from .object_id import Decode
from .pagination import pagination
from .response import Response
from .row_handler import Objects


def fixed_id_column(kwargs: dict | list | str):
    """Fixed the ID Column"""
    if isinstance(kwargs, dict):
        if kwargs.get("id"):
            kwargs["_id"] = kwargs["id"]
            del kwargs["id"]
    elif isinstance(kwargs, list):
        if "id" in kwargs:
            kwargs.remove("id")
            kwargs.append("_id")
    elif isinstance(kwargs, str):
        if kwargs == "id":
            kwargs = "_id"
        elif kwargs == "-id":
            kwargs = "-_id"
    return kwargs



"""
    Base
"""

import functools

from ..models import create_model
from .annotations import get_args
from .custom_class import create_custom_type
from .spoc import COMPONENT, component
from .table import table_info

# from .tools import get_module_name, to_pascal_case


def custom_type(
    original_object: object = None,
    *,
    table_name: str = None,
    primary_key: list = None,
    required: list = None,
    index: list = None,
    unique: list = None,
    unique_together: list = None,
    ignore: list = None,
    engine: str = "sql",
    database: str = "default",
    controller: object = None,
    description:str = None,
):
    """{ Controller } for GraphQL { Type } & { Model } Database"""
    primary_key = primary_key or []
    required = required or []
    index = index or []
    unique = unique or []
    unique_together = unique_together or []
    ignore = ignore or []

    config = {
        "table_name": table_name,
        "primary_key": primary_key,
        "required": required,
        "index": index,
        "unique": unique,
        "unique_together": unique_together,
        "ignore": ignore,
        "engine": engine,
        "database": database,
        "controller": controller,
    }

    # Starting Wrapper. . .
    if original_object is None:
        return functools.partial(
            custom_type,
            **config,
            description=description,
        )

    # Annotations (Python)
    model = get_args(original_object)

    # GraphQL (Strawberry) Class
    custom_class = create_custom_type(original_object, model, description)
    table_config = table_info(original_object, table_name, config)
    db_table_name = table_config.table_name
    config["table_name"] = db_table_name

    # Databases - (DBController)
    custom_class.__database__ = None
    if controller is not None:
        db_model = create_model(
            base=controller,
            new_object=custom_class,
            engine=engine,
            original_object=original_object,
            annotations=model.annotations.values(),
            config=config,
        )
        custom_class.__database__ = db_model

    # Component - (Spoc)
    component_config = {
        "table_name": db_table_name,
        "database": database,
        "engine": engine,
        "annotations": model.annotations,
    }
    component(custom_class, config=component_config, metadata=COMPONENT["model"])

    # Return: Customized Class
    return custom_class



import json
import pytest
from httpx import AsyncClient

from main import app

# Get GraphQL Operations
with open('client/main.graphql', 'r') as f:
    GRAPHQL = f.read()

# Create Client
async def client(operationName: str, variables: dict | None = None):
    """Create Client"""
    async with AsyncClient(app=app, base_url="http://127.0.0.1:8000") as ac:
        the_query = {"query": GRAPHQL, "operationName": operationName}
        if variables:
            the_query["variables"] = variables
        response = await ac.post("/graphql", json=the_query)
    return response

def check_output(expected, actual):
    """Check Output"""
    return json.dumps(actual, sort_keys = True, indent=0) == json.dumps(expected, sort_keys = True, indent=0)

@pytest.mark.asyncio
async def test_all_tasks():

    # API Request
    response = await client("AllTasks")

    # Expected Output
    expected_output = {'data': {'search': {'edges': [
        {
          "node": {
            "id": "Mjo6M2VmOWFiYmI1ZGY1YjY0MQ==",
            "title": "Post Tutorial",
            "description": "post the example app.",
            "status": "open"
          }
        },
        {
          "node": {
            "id": "MTo6YTU1ZTUzMmVhYjAyOGI0Mg==",
            "title": "First App",
            "description": "create an example app.",
            "status": "close"
          }
        }
    ]}}}
    
    # Assert
    assert response.status_code == 200
    assert check_output(response.json(), expected_output)
    

@pytest.mark.asyncio
async def test_search():

    # API Request
    variables = {"status": "open"}
    response = await client("SearchTask", variables)

    # Expected Output
    expected_output = {'data': {'search': {'edges': [
        {
            "node": {
            "id": "Mjo6M2VmOWFiYmI1ZGY1YjY0MQ==",
            "title": "Post Tutorial",
            "description": "post the example app.",
            "status": "open"
            }
        },
    ]}}}

    # Assert
    assert response.status_code == 200
    assert check_output(response.json(), expected_output)
    


"""
    Database - Table Config
"""
import dataclasses as dc

from ..tools import get_module_name


@dc.dataclass(frozen=True)
class TableConfig:
    """Table Setup"""

    table_uri: str = None
    table_name: str = None
    required: list = None
    primary_key: list = None
    index: list = None
    unique: list = None
    unique_together: list = None
    many_to_many: dict = None
    ignore: list = None
    engine: str = None
    database: str = None
    many_to_many: list = None


def table_info(original_object, table_name, config):
    """Table Info"""
    class_name = original_object.__name__
    class_module_name = get_module_name(original_object)
    if not class_module_name:
        class_module_name = "main"
    class_table_name = f"{class_module_name.lower()}_{class_name.lower()}"
    info_class_table_uri = f"{class_module_name.lower()}.{class_name.lower()}"
    if table_name:
        class_table_name = table_name
    # Set Name
    config["table_name"] = class_table_name
    config["table_uri"] = info_class_table_uri
    out_config = dict(config.items())
    del out_config["controller"]
    return TableConfig(**out_config)



"""
    Shell Functions
"""

import os
import pathlib
import shutil
import zipfile

TEMPORARY_DIR = pathlib.Path(__file__).parents[0] / "tmp"


def unzip_base(source, destination):
    """Unzip Base"""
    with zipfile.ZipFile(source, "r") as zip_ref:
        zip_ref.extractall(destination)


def unzip(source: pathlib.Path, destination: pathlib.Path):
    """Unzip Method"""
    zipfile_name = source.name.replace(".zip", "")
    unzip_base(source, TEMPORARY_DIR)
    tmp = TEMPORARY_DIR / zipfile_name
    file_names = os.listdir(tmp)
    for file_name in file_names:
        try:
            shutil.move(os.path.join(tmp, file_name), destination)
        except:
            pass
    shutil.rmtree(tmp, ignore_errors=False, onerror=None)



"""
    DataClass
"""
import dataclasses as dc
import typing


def pagination(toml):
    """Pagination"""
    max_items_per_page: int = toml.get("api", {}).get("items_per_page", 100)

    def init(self):
        page = max(1, self.page)
        limit = min(max_items_per_page, self.limit)
        if page < 0:
            page = 1
        self.page = page
        self.limit = max(limit, max_items_per_page)

    data_class = dc.make_dataclass(
        "Pagination",
        [
            ("page", typing.Optional[int], dc.field(default=1)),
            ("limit", typing.Optional[int], dc.field(default=100)),
        ],
        namespace={"__post_init__": init},
    )
    return data_class



"""
    Routers
"""
from fastapi import APIRouter


def routers(items: list):
    """Collect (FastAPI) Routers"""
    router = []
    for active in items:
        if isinstance(active.object, APIRouter):
            router.append(active.object)
    return router



"""
    Read & Transform the GraphQL Schema.
"""

import typing

SCALAR_TYPES = [
    "ID",
    "String",
    "Int",  # Integer
    "Float",
    "Boolean",
    "Decimal",
    "Date",
    "Datetime",
    "Time",
    "JSON",
]


class TypeModel(typing.TypedDict):
    """Object Info"""

    type: str
    name: str
    scalar: bool
    list: bool
    required: bool
    default: typing.Any


def set_default_value(setup):
    """Fix Default Value"""
    python_scalar = [
        "ID",
        "String",
        "Int",  # Integer
        "Float",
        "Boolean",
        "Decimal",
        "Date",
        "Datetime",
        "Time",
    ]
    if setup["default"] == "null":
        setup["default"] = None
    elif setup["scalar"] and setup["type"] in python_scalar:
        match setup["type"]:
            case "String":
                setup["default"] = str(setup["default"])
            case "Int":
                setup["default"] = int(setup["default"])
            case "Float":
                setup["default"] = float(setup["default"])
            case "Boolean":
                setup["default"] = setup["default"] == "true"
            case _:
                setup["default"] = str(setup["default"])
    return setup


def get_real_type(field):
    """Get Type from GraphQL Schema { Introspection }"""
    type_kind = field["type"]["kind"]
    type_name = field["type"]["name"]
    setup = None
    default_value = field.get("defaultValue") or "null"
    match type_kind:
        case "SCALAR":
            setup = TypeModel(
                type=field["type"]["name"],
                name=field.get("name"),
                scalar=True,
                list=False,
                required=False,
                default=default_value,
            )
        case "OBJECT":
            setup = TypeModel(
                type=type_name,
                name=field.get("name"),
                scalar=False,
                list=False,
                required=False,
                default=default_value,
            )
        case "LIST":
            child_type = field["type"]["ofType"]["ofType"]
            setup = TypeModel(
                type=child_type.get("name"),
                name=field.get("name"),
                scalar=child_type["kind"] == "SCALAR",
                list=True,
                required=False,
                default=default_value,
            )
        case "UNION":
            setup = TypeModel(
                type=type_name,
                name=field.get("name"),
                scalar=False,
                list=False,
                required=False,
                default=default_value,
            )
        case "NON_NULL":
            child_type = field["type"]["ofType"]
            match child_type["kind"]:
                case "SCALAR":
                    setup = TypeModel(
                        type=child_type.get("name"),
                        name=field.get("name"),
                        scalar=True,
                        list=False,
                        required=True,
                        default=default_value,
                    )
                case "OBJECT":
                    setup = TypeModel(
                        type=child_type.get("name"),
                        name=field.get("name"),
                        scalar=False,
                        list=False,
                        required=True,
                        default=default_value,
                    )
                case "LIST":
                    child_type = child_type["ofType"]
                    setup = TypeModel(
                        type=child_type.get("name"),
                        name=field.get("name"),
                        scalar=child_type["kind"] == "SCALAR",
                        list=True,
                        required=True,
                        default=default_value,
                    )
                case "UNION":
                    setup = TypeModel(
                        type=child_type.get("name"),
                        name=field.get("name"),
                        scalar=False,
                        list=False,
                        required=True,
                        default=default_value,
                    )
    setup = set_default_value(setup)
    return setup


def create_types(graphql_info):
    """Get Types from GraphQL Schema { Introspection }"""
    the_type = {}
    for active in graphql_info:
        the_fields = []
        for field in active["fields"]:
            the_fields.append(get_real_type(field))
        the_type[active["name"]] = the_fields
    return the_type


def create_inputs(graphql_info):
    """Get Inputs from GraphQL Schema { Introspection }"""
    the_type = {}
    for active in graphql_info:
        the_fields = []
        for field in active["fields"]:
            default_value = field.get("defaultValue") or "null"
            match field["type"]["kind"]:
                case "SCALAR":
                    setup = TypeModel(
                        type=field["type"]["name"],
                        name=field.get("name"),
                        scalar=True,
                        list=False,
                        required=False,
                        default=default_value,
                    )
                case "INPUT_OBJECT":
                    child_type = field["type"]
                    setup = TypeModel(
                        type=child_type.get("name"),
                        name=field.get("name"),
                        scalar=False,
                        list=False,
                        required=False,
                        default=default_value,
                    )
                case "LIST":
                    child_type = field["type"]["ofType"]
                    if child_type["kind"] == "NON_NULL":
                        child_type = child_type["ofType"]
                    setup = TypeModel(
                        type=child_type.get("name"),
                        name=field.get("name"),
                        scalar=child_type["kind"] == "SCALAR",
                        list=True,
                        required=True,
                        default=default_value,
                    )
                case "NON_NULL":
                    child_type = field["type"]["ofType"]
                    match child_type["kind"]:
                        case "INPUT_OBJECT":
                            setup = TypeModel(
                                type=child_type.get("name"),
                                name=field.get("name"),
                                scalar=False,
                                list=False,
                                required=True,
                                default=default_value,
                            )
                        case "LIST":
                            child_type = child_type["ofType"]
                            if child_type["kind"] == "NON_NULL":
                                child_type = child_type["ofType"]
                            setup = TypeModel(
                                type=child_type.get("name"),
                                name=field.get("name"),
                                scalar=child_type["kind"] == "SCALAR",
                                list=True,
                                required=True,
                                default=default_value,
                            )
                        case "SCALAR":
                            setup = TypeModel(
                                type=child_type.get("name"),
                                name=field.get("name"),
                                scalar=True,
                                list=False,
                                required=True,
                                default=default_value,
                            )
            setup = set_default_value(setup)
            the_fields.append(setup)
        the_type[active["name"]] = the_fields
    return the_type


def get_return_types(graphql_info):
    """API Return-Types"""
    responses = {}
    for item in graphql_info:
        real_type = item["response"]
        real_type = get_real_type({"type": real_type})
        # if real_type["type"].endswith("Connection"):
        #    real_type["type"] = real_type["type"].replace("Connection", "")
        del real_type["name"]
        del real_type["default"]
        responses[item["name"]] = real_type
    return responses


def get_info(schema):
    """Get Schema Info { Introspection }"""
    graphql_info = {
        "forms": [],
        "types": [],
        "ops": {
            "query": [],
            "mutation": [],
        },
        "returnTypes": {
            "query": [],
            "mutation": [],
        },
    }
    for item in schema["__schema"]["types"]:
        kind = item["kind"]
        match kind:
            case "INPUT_OBJECT":
                graphql_info["forms"].append(
                    {"name": item["name"], "fields": item["inputFields"]}
                )
            case "OBJECT":
                if item["name"] not in ["Query", "Mutation"] and not item[
                    "name"
                ].startswith("__"):
                    graphql_info["types"].append(
                        {"name": item["name"], "fields": item["fields"]}
                    )
                elif item["name"] in ["Query", "Mutation"]:
                    match item["name"]:
                        case "Query":
                            graphql_info["ops"]["query"] = [
                                {
                                    "name": active["name"],
                                    "fields": active["args"],
                                    "response": active["type"],
                                }
                                for active in item["fields"]
                            ]
                        case "Mutation":
                            graphql_info["ops"]["mutation"] = [
                                {
                                    "name": active["name"],
                                    "fields": active["args"],
                                    "response": active["type"],
                                }
                                for active in item["fields"]
                            ]

    # Custom Typing
    graphql_info["returnTypes"]["query"] = get_return_types(
        graphql_info["ops"]["query"]
    )
    graphql_info["returnTypes"]["mutation"] = get_return_types(
        graphql_info["ops"]["mutation"]
    )
    graphql_info["types"] = create_types(graphql_info["types"])
    graphql_info["forms"] = create_types(graphql_info["forms"])
    graphql_info["ops"]["query"] = create_inputs(graphql_info["ops"]["query"])
    graphql_info["ops"]["mutation"] = create_inputs(graphql_info["ops"]["mutation"])
    return graphql_info


# Introspection Query
INTROSPECTION_QUERY = """
query IntrospectionQuery {
	__schema {
		types {
			...FullType
		}
	}
}

fragment FullType on __Type {
	kind
	name
	fields(includeDeprecated: true) {
		name
		args {
			...InputValue
		}
		type {
			...TypeRef
		}
	}
	inputFields {
		...InputValue
	}
	interfaces {
		...TypeRef
	}
	enumValues(includeDeprecated: true) {
		name
		isDeprecated
		deprecationReason
	}
	possibleTypes {
		...TypeRef
	}
}

fragment InputValue on __InputValue {
	name
	type {
		...TypeRef
	}
	defaultValue
}

fragment TypeRef on __Type {
	kind
	name
	ofType {
		kind
		name
		ofType {
			kind
			name
			ofType {
				kind
				name
				ofType {
					kind
					name
					ofType {
						kind
						name
						ofType {
							kind
							name
							ofType {
								kind
								name
							}
						}
					}
				}
			}
		}
	}
}
"""



"""
    DataClass
"""
import dataclasses as dc
import typing


@dc.dataclass(frozen=True)
class FastAPI:
    """FastAPI"""

    title: str
    version: str
    description: typing.Any = None


@dc.dataclass(frozen=True)
class Extras:
    """API Extras"""

    middleware: typing.Any = None
    on_startup: typing.Any = None
    on_shutdown: typing.Any = None
    extensions: typing.Any = None
    permissions: typing.Any = None



"""
    Framework
"""
from types import SimpleNamespace

import spoc
from fastapi import FastAPI

from . import handlers

PLUGINS = ["types", "forms", "graphql", "router", "commands"]


def create_api(self):
    """FastAPI Builder"""

    app = FastAPI(
        title=self.info.title.title(),
        version=self.info.version,
        description=self.info.description,
    )
    # App Context
    self.context.app = app
    self.context.controller = self


@spoc.singleton
class Fastberry:
    """Framework"""

    def init(
        self,
    ):
        """Class __init__ Replacement"""
        framework = spoc.App(plugins=PLUGINS)

        # Self { Definitions }
        core_toml = framework.config["spoc"].get("spoc", {})

        self.settings = framework.settings
        self.context = SimpleNamespace(
            settings=framework.settings,
            is_production=core_toml.get("mode", "development") == "production",
        )

        # Info
        self.info = handlers.api_info(
            base_dir=framework.base_dir,
            pyproject=framework.config["pyproject"],
            toml=framework.config["spoc"],
        )

        # Create API
        create_api(self)
        self.core = framework
        self.extras = None
        self.router = None
        self.graphql = None
        self.toml = core_toml

        # Pagination : print(self.pagination(page=0, limit=500))
        self.pagination = handlers.pagination(core_toml)

        # FastAPI { Extras }
        self.extras = handlers.extras(framework.extras.items())

        if framework.component:
            # FastAPI { Routers }
            self.router = handlers.routers(framework.component.router.values())

            # GraphQL { Types }
            self.types = handlers.types(framework.component.types.values())

            # GraphQL { Forms }
            self.forms = handlers.forms(framework.component.forms.values())

            # GraphQL { Query & Mutation }
            self.graphql = handlers.graphql(
                schemas=framework.component.graphql.values(),
                permissions=self.extras.permissions,
            )

            # Command-Line-Interface
            self.cli = handlers.commands(framework.component.commands.values())

    def keys(self):
        """Finally: Collect { Keys }"""
        return sorted(
            [
                x
                for x in dir(self)
                if not x.startswith("_") and x not in ["init", "keys"]
            ]
        )


def graphql_schema():
    """GraphQL Schema

    Methods:
        * execute
        * execute_sync

    Example:

        schema.execute("graphql-query", variable_values={"key": "value"})

    Returns:
        schema: GraphQL-Schema
    """
    app = Fastberry()
    return app.graphql.schema()



"""
    Types
"""

try:
    import dbcontroller as dbc
except ImportError:
    dbc = None


def forms(models: list):
    """Collect (GraphQL) Input-Forms"""
    single_name: dict = {}
    all_forms: dict = {}
    if dbc:
        for active in models:
            is_component = dbc.is_form(active.object)
            if is_component:
                name = active.object.__name__
                found = single_name.get(name)
                if found and found.object != active.object:
                    app = found.app
                    err = f"""\n* <Form: { name }> is already in use by the <App: { app }>"""
                    err += """\n* <Forms> must have UNIQUE names."""
                    err += f"""\n* Rename it in either <App: { app } or { app }>."""
                    raise ValueError(err)
                single_name[name] = active
                all_forms[active.key] = active.object
    return all_forms



"""
    (Manager) - SQL & Mongo
"""
import functools
import typing

from .mongo import create_mongo_manager, mongo_model
from .sql import create_sql_manager, sql_model


def create_model(
    new_object: typing.Any = None,
    engine: str = None,
    original_object: typing.Any = None,
    annotations: typing.Any = None,
    config: typing.Any = None,
    base: typing.Any = None,
):
    """Create { SQL } or { Mongo }"""
    db_model = None
    match engine:
        case "sql":
            db_model = functools.partial(
                sql_model,
                controller=base,
                original_object=original_object,
                annotations=annotations,
                config=config,
                new_object=new_object,
            )
        case "mongo":
            db_model = functools.partial(
                mongo_model,
                controller=base,
                table_name=config["table_name"],
                annotations=annotations,
                config=config,
                new_object=new_object,
            )
    return db_model


def create_database(
    sql: typing.Any = None,
    mongo: typing.Any = None,
):
    """Create Database"""
    if sql:
        return create_sql_manager(sql)
    return create_mongo_manager(mongo)



# -*- coding: utf-8 -*-
"""
    { Settings }
"""
import pathlib

# Base Directory
BASE_DIR = pathlib.Path(__file__).parents[1]

# Installed Apps
INSTALLED_APPS = ["todo"]

# Database(s)
DATABASES = {
    "sql": {"default": "sqlite:///example.db"},
    "mongo": {"default": None},
}



"""
    Components
"""
import functools

import click
import spoc
from fastapi import APIRouter

COMPONENT = {}
COMPONENT["graphql"] = {"engine": "strawberry-graphql", "type": "schema"}
COMPONENT["router"] = {"engine": "fastapi", "type": "router"}
COMPONENT["commands"] = {"engine": "click", "type": "group"}

# Router
spoc.component(APIRouter, metadata=COMPONENT["router"])

# GraphQL Class @Decorator
def graphql(
    cls: object = None,
):
    """Strawberry { GraphQL } Creator"""
    spoc.component(cls, metadata=COMPONENT["graphql"])
    return cls


# Command-Line Interface
def cli(
    cls: object = None,
    *,
    group: bool = False,
):
    """Click { CLI } Creator"""
    if cls is None:
        return functools.partial(
            cli,
            group=group,
        )
    # Real Wrapper
    cls = click.group(cls)
    if not group:
        spoc.component(cls, metadata=COMPONENT["commands"])
    return cls



"""
    Create A New App
"""

import os.path

import click
import spoc

from .action.plugin_maker import create_plugin
from .shell import shell_print, unzip


@click.command()
@click.argument(
    "app_name",
    type=str,
    nargs=1,
)
# @click.option("--gql/--no-gql", default=False, type=bool, help="Init with sample code.")
@click.option(
    "--plugin/--no-plugin", default=False, type=bool, help="Init <App> as a plugin."
)
def start_app(app_name, plugin):
    """Creates a Fastberry App Directory."""
    # Get Path(s)
    apps_dir = spoc.base_dir / "apps"
    the_dir = apps_dir / app_name

    # Create Path(s)
    apps_dir.mkdir(parents=True, exist_ok=True)

    # Check Path(s)
    if os.path.isdir(the_dir):
        shell_print(
            f"""* Error: "{{ { app_name } }}" App Already Exists!""", color="red"
        )
    else:
        shell_print(
            f"""* Starting App: "{{ { app_name } }}" . . .""",
        )
        if plugin:
            the_dir.mkdir(parents=True, exist_ok=True)
            create_plugin(the_dir)
        else:
            unzip("base_app.zip", the_dir)



"""
    Built-in Forms for Fastberry(GraphQL).
"""
import fastberry as fb

try:
    import dbcontroller as dbc

    form = dbc.form.graphql
    value = dbc.form.field
    ID = dbc.ID
except ImportError:

    dbc = False


if dbc:
    core = form()

    ITEM_DOC = """
    Global **Selector** for **ITEM(s)** by **ID(s)**.
    """

    PAGINATION_DOC = """
    Global **Pagination**

    **Setting**: `(all: true)`

    **Returns**: **All** of the **items** in the **Database** (Mostly For **Admin(s) Use Only**).
    """

    @core(description=ITEM_DOC)
    class Item:
        """
        Item
        """

        id = value(
            ID,
            default=None,
        )
        ids = value(
            list[ID],
            default=None,
        )

        class Next:
            """Run After User-Input"""

            def run(self):
                """Run"""
                self.is_list = False
                if self.id is None and self.ids is not None:
                    self.is_list = True

    @core(description=PAGINATION_DOC)
    class Pagination:
        """
        Pagination
        """

        limit = value(
            int,
            default=10,
        )
        page = value(
            int,
            default=10,
        )
        sort_by = value(
            str,
            default="-id",
        )
        all = value(
            bool,
            default=False,
        )

        class Next:
            """Run After User-Input"""

            page = 0
            limit = 10

            def run(self):
                """Run"""
                controller = fb.App()
                page = controller.pagination(
                    page=self.page,
                    limit=self.limit,
                )
                self.page = page.page
                self.limit = page.limit

else:
    Item = type("Item", (object,), {})
    Pagination = type("Pagination", (object,), {})



from typing import Optional

from fastapi.concurrency import AsyncExitStack
from starlette.types import ASGIApp, Receive, Scope, Send


class AsyncExitStackMiddleware:
    def __init__(self, app: ASGIApp, context_name: str = "fastapi_astack") -> None:
        self.app = app
        self.context_name = context_name

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        if AsyncExitStack:
            dependency_exception: Optional[Exception] = None
            async with AsyncExitStack() as stack:
                scope[self.context_name] = stack
                try:
                    await self.app(scope, receive, send)
                except Exception as e:
                    dependency_exception = e
                    raise e
            if dependency_exception:
                # This exception was possibly handled by the dependency but it should
                # still bubble up so that the ServerErrorMiddleware can return a 500
                # or the ExceptionMiddleware can catch and handle any other exceptions
                raise dependency_exception
        else:
            await self.app(scope, receive, send)  # pragma: no cover



"""{ Global Values }
    Code for creating the core files in order to run the API.
"""

SETTINGS = '''
# -*- coding: utf-8 -*-
"""
    { Settings }
"""
import pathlib

# Base Directory
BASE_DIR = pathlib.Path(__file__).parents[1]

# Installed Apps
INSTALLED_APPS = []

# Database(s)
DATABASES = {
    "sql": {"default": None},  # Example: sqlite:///example.db
    "mongo": {"default": None},  # Example: mongodb://localhost:27017/example
}
'''.strip()

SPOC = """
[spoc]
mode = "custom"
custom_mode = "development"
docs = "config/docs.md"
generates = "graphql"

[spoc.api]
graphql_path = "/graphql"
max_depth = 4
items_per_page = 50
allowed_hosts = ["http://localhost", "http://localhost:8080"]

[spoc.apps]
production = []
development = []
staging = []

[spoc.extras]
middleware = []
extensions = []
permissions = []
on_startup = ["fastberry.extras.redirect_root"] # "fastberry.extras.graphql_info"
on_shutdown = []
""".strip()


DOCS = """
# Welcome

> This is a simple **API** Skeleton

---

## Links

> Go To [GraphQL](/graphql)

---

## Mode (Options)

- `development`

- `staging`

- `production`

---

## Settings Layout

```text
root/                           --> <Directory> - Project's Root.
|
|--  config/                    --> <Directory> - Configurations.
|    |
|    |-- .env/                  --> <Directory> - Environments.
|    |   |
|    |   |-- development.toml   --> <File> - Development    | Settings.
|    |   |-- production.toml    --> <File> - Production     | Settings.
|    |   `-- staging.toml       --> <File> - Staging        | Settings.
|    |
|    |-- docs.md                --> <File> - This Documentation is in HERE.
|    |-- settings.py            --> <File> - Python         | Settings.
|    `-- spoc.toml              --> <File> - TOML           | Settings.
|
|-- pyproject.toml              --> <File> - Project        | Settings.
|
`-- etc...
```
""".strip()



# Copyright 2009-present MongoDB, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Bits and pieces used by the driver that don't really fit elsewhere."""

import sys
import traceback
from collections import abc
from typing import Any, List, NoReturn

from bson.son import SON
from pymongo import ASCENDING
from pymongo.errors import (
    CursorNotFound,
    DuplicateKeyError,
    ExecutionTimeout,
    NotPrimaryError,
    OperationFailure,
    WriteConcernError,
    WriteError,
    WTimeoutError,
    _wtimeout_error,
)
from pymongo.hello import HelloCompat

# From the SDAM spec, the "node is shutting down" codes.
_SHUTDOWN_CODES = frozenset(
    [
        11600,  # InterruptedAtShutdown
        91,  # ShutdownInProgress
    ]
)
# From the SDAM spec, the "not primary" error codes are combined with the
# "node is recovering" error codes (of which the "node is shutting down"
# errors are a subset).
_NOT_PRIMARY_CODES: frozenset = (
    frozenset(
        [
            10058,  # LegacyNotPrimary <=3.2 "not primary" error code
            10107,  # NotWritablePrimary
            13435,  # NotPrimaryNoSecondaryOk
            11602,  # InterruptedDueToReplStateChange
            13436,  # NotPrimaryOrSecondary
            189,  # PrimarySteppedDown
        ]
    )
    | _SHUTDOWN_CODES
)
# From the retryable writes spec.
_RETRYABLE_ERROR_CODES: frozenset = _NOT_PRIMARY_CODES | frozenset(
    [
        7,  # HostNotFound
        6,  # HostUnreachable
        89,  # NetworkTimeout
        9001,  # SocketException
        262,  # ExceededTimeLimit
    ]
)


def _gen_index_name(keys):
    """Generate an index name from the set of fields it is over."""
    return "_".join(["%s_%s" % item for item in keys])


def _index_list(key_or_list, direction=None):
    """Helper to generate a list of (key, direction) pairs.

    Takes such a list, or a single key, or a single key and direction.
    """
    if direction is not None:
        return [(key_or_list, direction)]
    else:
        if isinstance(key_or_list, str):
            return [(key_or_list, ASCENDING)]
        if isinstance(key_or_list, abc.ItemsView):
            return list(key_or_list)
        elif not isinstance(key_or_list, (list, tuple)):
            raise TypeError("if no direction is specified, key_or_list must be an instance of list")
        return key_or_list


def _index_document(index_list):
    """Helper to generate an index specifying document.

    Takes a list of (key, direction) pairs.
    """
    if isinstance(index_list, abc.Mapping):
        raise TypeError(
            "passing a dict to sort/create_index/hint is not "
            "allowed - use a list of tuples instead. did you "
            "mean %r?" % list(index_list.items())
        )
    elif not isinstance(index_list, (list, tuple)):
        raise TypeError("must use a list of (key, direction) pairs, not: " + repr(index_list))
    if not len(index_list):
        raise ValueError("key_or_list must not be the empty list")

    index: SON[str, Any] = SON()
    for (key, value) in index_list:
        if not isinstance(key, str):
            raise TypeError("first item in each key pair must be an instance of str")
        if not isinstance(value, (str, int, abc.Mapping)):
            raise TypeError(
                "second item in each key pair must be 1, -1, "
                "'2d', or another valid MongoDB index specifier."
            )
        index[key] = value
    return index


def _check_command_response(
    response, max_wire_version, allowable_errors=None, parse_write_concern_error=False
):
    """Check the response to a command for errors."""
    if "ok" not in response:
        # Server didn't recognize our message as a command.
        raise OperationFailure(
            response.get("$err"), response.get("code"), response, max_wire_version
        )

    if parse_write_concern_error and "writeConcernError" in response:
        _error = response["writeConcernError"]
        _labels = response.get("errorLabels")
        if _labels:
            _error.update({"errorLabels": _labels})
        _raise_write_concern_error(_error)

    if response["ok"]:
        return

    details = response
    # Mongos returns the error details in a 'raw' object
    # for some errors.
    if "raw" in response:
        for shard in response["raw"].values():
            # Grab the first non-empty raw error from a shard.
            if shard.get("errmsg") and not shard.get("ok"):
                details = shard
                break

    errmsg = details["errmsg"]
    code = details.get("code")

    # For allowable errors, only check for error messages when the code is not
    # included.
    if allowable_errors:
        if code is not None:
            if code in allowable_errors:
                return
        elif errmsg in allowable_errors:
            return

    # Server is "not primary" or "recovering"
    if code is not None:
        if code in _NOT_PRIMARY_CODES:
            raise NotPrimaryError(errmsg, response)
    elif HelloCompat.LEGACY_ERROR in errmsg or "node is recovering" in errmsg:
        raise NotPrimaryError(errmsg, response)

    # Other errors
    # findAndModify with upsert can raise duplicate key error
    if code in (11000, 11001, 12582):
        raise DuplicateKeyError(errmsg, code, response, max_wire_version)
    elif code == 50:
        raise ExecutionTimeout(errmsg, code, response, max_wire_version)
    elif code == 43:
        raise CursorNotFound(errmsg, code, response, max_wire_version)

    raise OperationFailure(errmsg, code, response, max_wire_version)


def _raise_last_write_error(write_errors: List[Any]) -> NoReturn:
    # If the last batch had multiple errors only report
    # the last error to emulate continue_on_error.
    error = write_errors[-1]
    if error.get("code") == 11000:
        raise DuplicateKeyError(error.get("errmsg"), 11000, error)
    raise WriteError(error.get("errmsg"), error.get("code"), error)


def _raise_write_concern_error(error: Any) -> NoReturn:
    if _wtimeout_error(error):
        # Make sure we raise WTimeoutError
        raise WTimeoutError(error.get("errmsg"), error.get("code"), error)
    raise WriteConcernError(error.get("errmsg"), error.get("code"), error)


def _get_wce_doc(result):
    """Return the writeConcernError or None."""
    wce = result.get("writeConcernError")
    if wce:
        # The server reports errorLabels at the top level but it's more
        # convenient to attach it to the writeConcernError doc itself.
        error_labels = result.get("errorLabels")
        if error_labels:
            wce["errorLabels"] = error_labels
    return wce


def _check_write_command_response(result):
    """Backward compatibility helper for write command error handling."""
    # Prefer write errors over write concern errors
    write_errors = result.get("writeErrors")
    if write_errors:
        _raise_last_write_error(write_errors)

    wce = _get_wce_doc(result)
    if wce:
        _raise_write_concern_error(wce)


def _fields_list_to_dict(fields, option_name):
    """Takes a sequence of field names and returns a matching dictionary.

    ["a", "b"] becomes {"a": 1, "b": 1}

    and

    ["a.b.c", "d", "a.c"] becomes {"a.b.c": 1, "d": 1, "a.c": 1}
    """
    if isinstance(fields, abc.Mapping):
        return fields

    if isinstance(fields, (abc.Sequence, abc.Set)):
        if not all(isinstance(field, str) for field in fields):
            raise TypeError(
                "%s must be a list of key names, each an instance of str" % (option_name,)
            )
        return dict.fromkeys(fields, 1)

    raise TypeError("%s must be a mapping or list of key names" % (option_name,))


def _handle_exception():
    """Print exceptions raised by subscribers to stderr."""
    # Heavily influenced by logging.Handler.handleError.

    # See note here:
    # https://docs.python.org/3.4/library/sys.html#sys.__stderr__
    if sys.stderr:
        einfo = sys.exc_info()
        try:
            traceback.print_exception(einfo[0], einfo[1], einfo[2], None, sys.stderr)
        except IOError:
            pass
        finally:
            del einfo



"""
    * PAGINATION
"""
from collections import namedtuple

# Page
Page = namedtuple("Page", ["offset", "limit"])


def pagination(page: int = 1, limit: int = 100):
    """[summary]
    Database Pagination.
    """
    offset = (page - 1) * limit
    return Page(offset, limit)
